package client

import (
	"context"
	"crypto/tls"
	"crypto/x509"
	"errors"
	"fmt"
	"net/url"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/cache"
	clientsSvc "git.perx.ru/perxis/perxis-go/pkg/clients/middleware"
	clientsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/clients/transport/grpc"
	collaboratorsSvc "git.perx.ru/perxis/perxis-go/pkg/collaborators/middleware"
	collaboratorsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/collaborators/transport/grpc"
	collectionsSvc "git.perx.ru/perxis/perxis-go/pkg/collections/middleware"
	collectionsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/collections/transport/grpc"
	"git.perx.ru/perxis/perxis-go/pkg/content"
	environmentsSvc "git.perx.ru/perxis/perxis-go/pkg/environments/middleware"
	environmentsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/environments/transport/grpc"
	invitationsSvc "git.perx.ru/perxis/perxis-go/pkg/invitations/middleware"
	invitationsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/invitations/transport/grpc"
	itemsSvc "git.perx.ru/perxis/perxis-go/pkg/items/middleware"
	itemsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/items/transport/grpc"
	localesSvc "git.perx.ru/perxis/perxis-go/pkg/locales/middleware"
	localsTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/locales/transport/grpc"
	referencesSvc "git.perx.ru/perxis/perxis-go/pkg/references/middleware"
	referencesTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/references/transport/grpc"
	rolesSvc "git.perx.ru/perxis/perxis-go/pkg/roles/middleware"
	rolesTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/roles/transport/grpc"
	spacesSvc "git.perx.ru/perxis/perxis-go/pkg/spaces/middleware"
	spacesTransportGrpc "git.perx.ru/perxis/perxis-go/pkg/spaces/transport/grpc"
	"go.uber.org/zap"
	"golang.org/x/oauth2/clientcredentials"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/grpc/credentials/oauth"
	"google.golang.org/grpc/metadata"
)

const (
	DefaultCacheSize = 1000
	DefaultCacheTTL  = time.Second * 10
)

func NewClient(addr string, opts ...Option) (*content.Content, *grpc.ClientConn, error) {
	ctx := context.Background()

	client := &content.Content{}
	dialOpts := make([]grpc.DialOption, 0)
	config := &Config{}

	for _, o := range opts {
		o(config)
	}

	if config.Logger == nil {
		config.Logger = zap.NewNop()
	}

	authDialOpts, err := config.GetAuthDialOpts(ctx)
	if err != nil {
		return nil, nil, err
	}
	dialOpts = append(dialOpts, authDialOpts...)

	contentConn, err := grpc.Dial(addr, dialOpts...)
	if err != nil {
		return nil, nil, err
	}

	client.Spaces = spacesTransportGrpc.NewClient(contentConn, config.ClientOptions...)
	client.Environments = environmentsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Collections = collectionsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Items = itemsTransportGrpc.NewClient(contentConn, config.ClientOptions...)
	client.Invitations = invitationsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Collaborators = collaboratorsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Clients = clientsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Locales = localsTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.Roles = rolesTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)
	client.References = referencesTransportGrpc.NewGRPCClient(contentConn, "", config.ClientOptions...)

	if !config.NoDecode {
		client.Items = itemsSvc.ClientEncodeMiddleware(client.Collections)(client.Items)
		client.References = referencesSvc.ClientEncodeMiddleware(client.Collections)(client.References)
	}

	if !config.NoCache {
		client = WithCaching(client, DefaultCacheSize, DefaultCacheTTL)
	}

	if !config.NoLog {
		client = WithLogging(client, config.Logger, config.AccessLog)
	}

	return client, contentConn, nil
}

func WithCaching(client *content.Content, size int, ttl time.Duration) *content.Content {
	c := *client

	c.Clients = clientsSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Clients)
	c.Environments = environmentsSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Environments)
	c.Locales = localesSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Locales)
	c.Roles = rolesSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Roles)
	c.Spaces = spacesSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Spaces)
	c.Items = itemsSvc.CachingMiddleware(cache.NewCache(size, ttl), cache.NewCache(size, ttl), c.Environments)(client.Items)
	c.Collections = collectionsSvc.CachingMiddleware(cache.NewCache(size, ttl), c.Environments)(client.Collections)
	c.Collaborators = collaboratorsSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Collaborators)
	c.Invitations = invitationsSvc.CachingMiddleware(cache.NewCache(size, ttl))(client.Invitations)

	return &c
}

func WithLogging(cs *content.Content, logger *zap.Logger, accessLog bool) *content.Content {
	s := *cs

	s.Collaborators = collaboratorsSvc.WithLog(s.Collaborators, logger, accessLog)
	s.Collections = collectionsSvc.WithLog(s.Collections, logger, accessLog)
	s.Environments = environmentsSvc.WithLog(s.Environments, logger, accessLog)
	s.Invitations = invitationsSvc.WithLog(s.Invitations, logger, accessLog)
	s.Items = itemsSvc.WithLog(s.Items, logger, accessLog)
	s.Locales = localesSvc.WithLog(s.Locales, logger, accessLog)
	s.Roles = rolesSvc.WithLog(s.Roles, logger, accessLog)
	s.Spaces = spacesSvc.WithLog(s.Spaces, logger, accessLog)
	s.Clients = clientsSvc.WithLog(s.Clients, logger, accessLog)
	s.References = referencesSvc.WithLog(s.References, logger, accessLog)

	return &s
}

func AddAPIKeyInterceptor(key string) grpc.UnaryClientInterceptor {
	return func(ctx context.Context, method string, req, reply interface{}, cc *grpc.ClientConn, invoker grpc.UnaryInvoker, opts ...grpc.CallOption) error {
		ctx = metadata.AppendToOutgoingContext(ctx, "authorization", "API-Key"+" "+key)
		return invoker(ctx, method, req, reply, cc, opts...)
	}
}

func (c *Config) GetAuthDialOpts(ctx context.Context) (opts []grpc.DialOption, err error) {

	if c.Auth == nil {
		opts = append(opts, grpc.WithTransportCredentials(insecure.NewCredentials()))
		return
	}

	authConfig := c.Auth

	switch {
	case authConfig.OAuth2 != nil:
		// create external grpc client
		conf := &clientcredentials.Config{
			TokenURL:       authConfig.OAuth2.TokenURL,
			ClientID:       authConfig.OAuth2.ClientID,
			ClientSecret:   authConfig.OAuth2.ClientSecret,
			EndpointParams: url.Values{"audience": {authConfig.OAuth2.Audience}},
		}

		// обязательно использовать tls для credentials oauth.TokenSource https://github.com/grpc/grpc-go/blob/64031cbfcf4d84c026be93ad7b74b3c290100893/credentials/oauth/oauth.go#L160
		if authConfig.Insecure {
			return nil, errors.New("oauth requires tls")
		}
		if authConfig.TLS == nil {
			authConfig.TLS = &TLS{SkipVerify: true}
		}

		opts = append(opts,
			grpc.WithPerRPCCredentials(oauth.TokenSource{
				TokenSource: conf.TokenSource(ctx),
			}),
		)
	case authConfig.APIKey != nil:

		if !authConfig.Insecure && authConfig.TLS == nil {
			authConfig.TLS = &TLS{SkipVerify: true}
		}
		opts = append(opts,
			grpc.WithUnaryInterceptor(AddAPIKeyInterceptor(authConfig.APIKey.APIKey)),
		)
	}

	switch {
	case authConfig.TLS != nil && !authConfig.SkipVerify:
		certPool := x509.NewCertPool()
		if !certPool.AppendCertsFromPEM(authConfig.TLS.CaCert) {
			return nil, fmt.Errorf("CA certificate not loaded")
		}

		clientCert, err := tls.X509KeyPair(authConfig.TLS.Cert, authConfig.TLS.Key)
		if err != nil {
			return nil, err
		}

		opts = append(opts,
			grpc.WithTransportCredentials(credentials.NewTLS(&tls.Config{
				Certificates: []tls.Certificate{clientCert},
				RootCAs:      certPool,
			})),
		)
	case authConfig.TLS != nil && authConfig.SkipVerify:
		opts = append(opts, grpc.WithTransportCredentials(credentials.NewTLS(&tls.Config{InsecureSkipVerify: true})))
	default:
		opts = append(opts, grpc.WithTransportCredentials(insecure.NewCredentials()))
	}

	return
}
